(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(*
main.sml: The main module of Level 0/1 prover. It contains the user interface
for loading and executing Level 0/1 programs.
*)

signature MAIN =
sig
	structure Core : CORE
	structure Prover0 : PROVER0
	structure Prover1 : PROVER1
	
	exception UserExits

	val go : string -> unit
	val golist : string list -> unit
	val select_prover : int -> unit
	val set_path : string -> unit
end

structure Main : MAIN =
struct

structure Norm = NormFunc(structure Term = Term)
structure LamTree = FuncLamTree(structure Norm = Norm structure Term = Term)
structure Unif = HO_Pattern_Unify(structure Term = Norm.Term  structure Norm = Norm)
structure Sub = SubstFunc(structure Term = Norm.Term structure Norm = Norm)
structure Cls = ClauseFunc(structure Tm = Term structure Sb = Sub structure Norm = Norm)
structure PPrint = PPrintFunc(
			structure Term = Term
			structure Norm = Norm
			structure Sub = Sub
			structure UStream = UStream
		)

(* Parser interface *)					
structure ParInt = PIntFunct(structure Term = Term
				structure Cls = Cls)

structure Lam = LamFunc(structure Term = Term 
			structure Cls = Cls 
			structure ParInt = ParInt)

structure Core = CoreFunc(
	structure Term = Term
	structure Norm = Norm
	structure LamTree = LamTree
	structure Cls = Cls
	structure Sub = Sub
	structure Unif = Unif
	structure UStream = UStream
	structure Lam = Lam
	structure ParInt = ParInt
	structure PPrint = PPrint
)

structure Prover0 = Prover0Func(structure Core = Core)
structure Prover1 = Prover1Func(structure Core = Core 
			structure Prover0 = Prover0)

exception UserExits

fun set_path S = Core.set_path S

open Core
val prove = ref (Prover1.prove)


fun select_prover 0 = (prove := Prover0.prove)
  | select_prover 1 = (prove := Prover1.prove)
  | select_prover n = (print "Invalid choice. Prover not changed.\n")

fun interp(prog) =
	(
	  (* const_level := 0; *)
  	  Term.reset_counter();
	  let val p = (Lam.reset_line(); Lam.getQuery())
				handle _ => (raise InvalidGoal)     		  
    	  in
    	    case p of
    	       	(SOME (ParInt.ParseGoal goal)) =>
    	       		let val vars = rev (Term.freevars [goal])
     				in 
     					PPrint.present_substs(
     					   (!prove)(0, prog,[ref (Term.app(judg, [goal]))],Sub.empty),vars)
     				end
     		  | (SOME (ParInt.ParseCommand ParInt.exit)) => 
     		  		raise UserExits
     		  | _ => (print "Invalid goal.\n")	
          end   		  
   	)

val VarPrefixes = ref ("","","")
   	
fun loop(prog) =
    (
       print "\n?- "; 
       VarPrefixes := Term.get_prefixes();
       (
       	interp(prog) 
        	handle LrParser.ParseError => (print "Parse error.\n")
            	| Term.NotValidTerm => (print "Error: invalid term.\n")
            	| Sub.NotPureTerm => (print "Error: non-pure term found in implicational goal. \n")
            	| Unif.NotLLambda => (print "Error: non L-lambda unification pair found. \n")            
            	| InvalidGoal => (print "Error: invalid goal.\n")
       );
       Term.set_prefixes (!VarPrefixes);
       loop(prog)
    )

fun go file = 
    let val prog = (Term.reset_counter(); Term.reset_name();
                    ParInt.reset_symTbl(); Lam.loadFile(!path, file))
    in
         print "Type `exit.' to exit.\n";
         loop(prog)
         handle UserExits => (print "Goodbye...\n")
    end

fun golist fs = 
	let val prog = (Term.reset_counter(); Term.reset_name();
			ParInt.reset_symTbl(); Lam.loadFiles(!path, fs))
	in
		print "Type `exit.' to exit.\n"; 
		loop(prog)
		handle UserExits => (print "Goodbye...\n")
	end
          
     
end (* of structure *)
