(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* Interface to the lambda term parser *)

signature LAM =
sig
   structure ParInt : PARSERINT
   structure Cls : CLAUSE
   structure Term : TERM
   sharing Cls.Term = Term and ParInt.Term = Term and ParInt.Cls = Cls

   val getQuery : unit -> ParInt.select option 
   val load : TextIO.instream -> Cls.clause list   
   val loadFile : string * string -> Cls.clause list
   val loadFiles : string * (string list) -> Cls.clause list 
   val reset_line : unit -> unit
end

functor LamFunc(structure Term : TERM 
		structure Cls : CLAUSE 
		structure ParInt : PARSERINT
		sharing Cls.Term = Term and 
		ParInt.Term = Term and Cls = ParInt.Cls) : LAM =
struct

structure Term = Term
structure Cls = Cls
structure ParInt = ParInt
structure Basics = Basics

structure LamLrVals = LamLrValsFun(
				structure Token = LrParser.Token
				structure Term = Term
				structure Cls = Cls
				structure ParInt = ParInt)

structure LamLex = LamLexFun(structure Tokens = LamLrVals.Tokens)

structure LamParser =
   Join(structure LrParser = LrParser
        structure ParserData = LamLrVals.ParserData
        structure Lex = LamLex);

val truec = ref (Term.const("true", 0, Term.none))

fun reset_line() = (LamLex.UserDeclarations.pos := 0)

fun invoke lexstream =
    let fun print_error (s, i:int, _) =
          TextIO.output(TextIO.stdOut,
          "Error, line " ^ (Int.toString i) ^ ", " ^ s ^ "\n")
    in 
      LamParser.parse(0,lexstream,print_error, ())
    end

fun load instream =
	let val lexer = LamParser.makeLexer
                          (fn _ =>
                              let val S = TextIO.inputLine instream
                              in
                                  case S of (SOME t) => t
                                          | NONE => ("")
                              end
                           )
        val dummyEOF = LamLrVals.Tokens.EOF(0,0)
        fun loop lexer L =
            let val (result, lexer) = invoke lexer
                val (nextToken,lexer) = LamParser.Stream.get lexer       
	    in
	        case result of 
				SOME (ParInt.ParseClause cl) =>
                		(if LamParser.sameToken(nextToken,dummyEOF) then 
                			(print "EOF!\n"; (cl::L))
                		else loop lexer  (cl::L)
						)
		     |  SOME (ParInt.ParseGoal G) => 
					let val cl = Cls.makeClause(G, truec, Term.freevars [G])
					in
                		(if LamParser.sameToken(nextToken,dummyEOF) then 
                				(print "EOF!\n"; (cl::L))
                			else loop lexer  (cl::L)
						)					
					end
		     |  NONE => L
		     |  SOME _ => (print "Invalid definition clause\n"; L)
            end	    
     in rev(loop lexer nil) end

fun getQuery() =
    let val lexer = LamParser.makeLexer
                          (fn _ =>
                              let val S = TextIO.inputLine (TextIO.stdIn)
                              in
                                  case S of (SOME t) => t
                                          | NONE => ("")
                              end
                           )
        val (result, lexer) = invoke lexer
    in
		result
    end    


val filelist : (string list) ref = ref nil

fun loadfs(path, fname) =
   let val file = path^fname
   in                 
   if (Basics.member(file, !filelist)) then nil
   else
   (
        filelist := file::(!filelist) ;
	let val ins = (SOME (TextIO.openIn(file)))
		handle IO.Io(_) => 
			(print "Failed in opening file "; print file; 
			 print "\n"; NONE)	
	in
	   case ins of (SOME instream) =>
		(
			let val lexer = LamParser.makeLexer
                          		(fn _ =>
                              			let val S = TextIO.inputLine instream
                              			in
                                  		case S of (SOME t) => t
                                          	| NONE => ("")
                              			end
                           		)
        		   val dummyEOF = LamLrVals.Tokens.EOF(0,0)
        		   fun loop lexer L =
            		      let val (result, lexer) = invoke lexer
                		  val (nextToken,lexer) = LamParser.Stream.get lexer       
	    		      in
	        		case result of SOME (ParInt.ParseClause cl) =>
                		   if LamParser.sameToken(nextToken,dummyEOF) then 
                			(print "EOF!\n"; (cl::L))
                		   else loop lexer  (cl::L)

		                 |  SOME (ParInt.ParseGoal G) => 
					let val cl = Cls.makeClause(G, truec, Term.freevars [G])
					in
                		            if LamParser.sameToken(nextToken,dummyEOF) then 
                				(print "EOF!\n"; (cl::L))
                		 	     else loop lexer  (cl::L)
					end
		                 |  NONE => L
		                 |  SOME (ParInt.ParsePreamble (ParInt.incl ifile)) => 
                                     (
                                       print "Include file : ";
                                       print (path^ifile); print "\n";
                                       loop lexer ( (rev (loadfs(path, ifile))) @ L)
                                     )
                                 |  _ => (print "Invalid clause\n."; L)
                              end	    
                            val res = rev(loop lexer nil) 
                                      handle E => (TextIO.closeIn(instream); raise E)
                         in (TextIO.closeIn(instream); res) end
		)
           | NONE => nil
	end
   )
   end
   
fun loadFile(path, file) = 
  let val fpath = if (path = "" orelse (String.isSuffix "/" path) orelse 
                      (String.isSuffix "\\" path)) then path
                  else path^"/"
  in                  
 	 filelist := nil; loadfs(fpath, file)
  end

fun loadFiles(path, files) =
  let val fpath = if (path = "" orelse (String.isSuffix "/" path) orelse 
                      (String.isSuffix "\\" path)) then path
                  else path^"/"
  in                  
 	 filelist := nil; 
 	 foldr (fn (file, l) => loadfs(fpath, file) @ l) nil files
  end
    
end (* of functor *)


