(****************************************************************************)
(* An implementation of Higher-Order Pattern Unification in SML             *)
(* Copyright (C) 2004 Gopalan Nadathur and Natalie Linnell                  *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* Modified by Alwen Tiu, September 2004.                                   *)

(* 
unify.sml: This module implements the higher-order pattern unification function
by Gopalan Nadathur and Natalie Linnel.
*)

signature HO_PATTERN_UNIFY = 
sig 

structure Term : TERM

exception UniEx of string
exception NotLLambda
exception ConstClash of (Term.term * Term.term)

val pattern_unify : Term.term -> Term.term -> unit
end


functor HO_Pattern_Unify(structure Term : TERM
                         structure Norm : NORM
                         sharing Norm.Term = Term) : HO_PATTERN_UNIFY =
struct

structure Term = Term

(*
exception OccursCheck
exception ArgMismatch
exception TimeStampInconsistent
exception ConstClash
exception Other
exception TypesMismatch
*)

exception UniEx of string

val OccursCheck = UniEx("OccursCheck")
val ArgMismatch = UniEx("ArgMismatch")
val TimeStampInconsistent = UniEx("TimeStampInconsistent")
(* val ConstClash = UniEx("ConstClash") *)

val Other = UniEx("Other")
val TypesMismatch = UniEx("TypesMismatch")
(* val NotLLambda = UniEx("NotLLambda") *)

exception NotLLambda
exception ConstClash of (Term.term * Term.term)


				   
(* transforming a term to represent substitution under abstractions *)
fun lift (c as (ref (Term.const _))) n = c
  | lift (v as (ref (Term.var _))) n = v
  | lift (ref (Term.dB i)) n = (ref (Term.dB (i+n)))
  | lift t n = ref (Term.susp (t,0,n,nil))


(* transforming a list of arguments to represent eta fluffing *)
fun lift_args nil n = 
      if (n = 0) then nil else ((ref (Term.dB n))::lift_args nil (n-1))
  | lift_args (a::rargs) n = ((lift a n)::(lift_args rargs n))

(* Check if a constant appears in a list of terms *)
fun unique_const _ nil = true
  | unique_const c ((ref (Term.const c'))::rargs) =
        if (c = c') then false else unique_const c rargs
  | unique_const c (_::rargs) = unique_const c rargs

(* Check if a bound variable appears in a list of terms *)
fun unique_bv _ nil = true
  | unique_bv i ((ref (Term.dB j))::rargs) = 
        if (i = j) then false else (unique_bv i rargs)
  | unique_bv i (_::rargs) = unique_bv i rargs

(* Check if a list of terms meets the LLambda requirements for the
   arguments of a flexible term. The second argument is the time stamp
   of the variable head. 
*)
fun check_flex_args nil _ = true
  | check_flex_args ((ref (Term.const (c' as (c,cts,_))))::rargs) fts =
      if ((cts > fts) andalso (unique_const c' rargs)) 
      then check_flex_args rargs fts
      else false
  | check_flex_args ((ref (Term.dB i))::rargs) fts =
      if (unique_bv i rargs) then check_flex_args rargs fts
      else false
  | check_flex_args (_::rargs) _ = false


(* return a nonzero index if a bound variable appears in a list of
   arguments; the index is the position from the right, representing
   the de Bruijn index of the abstraction capturing the argument. Zero
   signals absence in the argument list. Arguments in the list are
   expected to be (head) normalized and dereferenced
*)
fun bvindex (_,nil,_) = 0
  | bvindex (bv as Term.dB(i), (ref (Term.dB j) :: rargs), n) = 
      if (i = j) then n else bvindex (bv,rargs,n-1)
  | bvindex(bv,(_::rargs),n) = bvindex (bv,rargs,n-1)


(* return a nonzero index if a constant appears in a list of
   arguments; the index is the position from the right, representing
   the de Bruijn index of the abstraction capturing the argument. Zero
   signals absence in the argument list. Arguments in the list are
   expected to be (head) normalized and dereferenced
*)
fun cindex (_,nil,_) = 0
  | cindex (c1 as (Term.const _),(ref(c2 as (Term.const _))::rargs),n) =
      if (c1 = c2) then n else cindex (c1,rargs,n-1)
  | cindex (c,(_::rargs),n) = cindex (c,rargs,n-1) 


(* Given a flexible term (v1 a11 ... a1n) and another term of the form 
   (... (v2 a21 ... a2m) ...) where v1 and v2 are distinct variables,
   ts1 and ts2 are the timestamps associated with v1 and v2, a1 and a2
   represent the sequence a11 ... a1n and a21 ... a2m and lev
   represents the number of abstractions under which v2 appears
   embedded in the second term, return a triple consisting of 
     (i) a truth value indicating whether a pruning or raising
         substitution is needed for v2,
    (ii) a list of terms b1 ... bk such that the term 
            lam ... lam (... (v2' b1 ... bk) ...) 
         represents a unifying substitution for v1; these terms 
         consist of constants from a11 ... a1n over which v2 is
         possibly raised and inversions of a pruned a21 ... a2m, and
   (iii) the arguments c1 ... ck of a possible "raising" and pruning
         substitution for v2 matching the arguments b1 ... bk for
         v2' in (ii).
   The composition of the arguments lists can be understood
   qualitatively as follows: If (ts1 < ts2) then the initial part of
   b1 ... bk is the indices of constants from a11 ... a1n that do
   not appear in a21 ... a2m and that have a timestamp less than or 
   equal to ts2 (this corresponds to raising v2) and the rest of 
   b1 ... bk are the indices (relative to the list a11 ... a1n) of 
   the constants in a21 ... a2m that appear in a11 ... a1n (these are 
   the arguments that must not be pruned). Correspondingly, the first
   part of the list c1 ... ck are the constants from a11 ... a1n over
   which v2 is "raised" and the second part are the indices relative
   to a21 ... a2m of the constants that are not pruned. If (ts1 >= ts2)
   then each of b1 ... bk is either a constant in a21 ... a2m that
   does not appear in a11 ... a1n and which has a timestamp less than
   ts1 (this corresponds to first raising v1 and then reducing the
   substitution generated for v1) or it is the index, relative to 
   a11 ... a1n, of the terms in a21 ... a2m that are in 
   a11 ... a1n. The list c1 ... ck in this case are simply the indices
   relative to a21 ... a2m of the terms in a21 ... a2m that are
   preserved (i.e. not pruned) in b1 ... bk.

   This definition assumes that the terms in a1 and a2 are in
   head-normal and dereferenced form and that a1 satisfies the LLambda 
   requirements. If a2 does not satisfy these requirements, an
   exception will be raised.  
*)
fun raise_and_invert ts1 ts2 a1 a2 lev =
      let 

        val l1 = (length a1) + lev
 
        (* (raise args n) generates the collection of constants in as
           that have a time stamp < ts2 and the de Bruijn indices,
           assuming n is the index for the abstraction corresponding
           to the first term in args, for these constants. This
           serves to raise v2 in the case that ts1 < ts2. The boolean
           component of the triple returned is set to true if there is
           any raising to be done. The terms in args are assumed to be
           constants or de Bruijn indices as head normalized and
           dereferenced arguments satisfying the LLambda restriction
           would be. 
        *) 
        fun raise_var nil _ = (false,nil,nil)
          | raise_var ((ref (Term.dB _))::rargs) n = raise_var rargs (n-1)
          | raise_var ((c as ref (Term.const (_,cts,_)))::rargs) n =
              let val rest as (raised,inds,consts) = (raise_var rargs (n-1))
              in if (cts <= ts2) 
                 then (true,(ref (Term.dB (n+lev)))::inds,c::consts) 
                 else rest
              end

        (* (prune args n) "prunes" those items in args that are not
           bound by an embedded abstraction and that do not appear in
           a1 and inverts those items that are not pruned and that are
           not bound by an embedded abstraction; n is assumed to be
           the length of args here and hence yields the index of the
           leftmost argument position. This pruning computation is
           relevant to the case when (ts1 < ts2). The terms in args
           are assumed to be constants or de Bruijn indices. 
        *)
        fun prune nil 0 = (false,nil,nil)
          | prune ((a as (ref (Term.dB i)))::rargs) n = 
              let val (pruned,inds1,inds2) = prune rargs (n-1)
              in if (i > lev) 
                 then let val j = bvindex((Term.dB (i-lev)),a1,l1)
                      in if (j = 0) 
                         then (true,inds1,inds2)
                         else (pruned,
                               (ref (Term.dB (j+lev)))::inds1,
                               (ref (Term.dB n))::inds2)
                      end
                 else (pruned,a::inds1,(ref (Term.dB n))::inds2)
              end
          | prune ((ref (c as Term.const _))::rargs) n =
              let val (pruned,inds1,inds2) = prune rargs (n-1)
                  val j = cindex(c,a1,l1)
              in if (j = 0)
                 then (true,inds1,inds2)
                 else (pruned,
                       (ref (Term.dB (j+lev)))::inds1,
                       (ref (Term.dB n))::inds2)
              end

        (* Revelant to the case when ts1 > ts2. In this case,
           (prune_and_raise args n) prunes those constants and de
           Bruijn indices not bound by an embedded abstraction that do
           not appear in a1 and, in the case of constants, that do not
           have a timestamp < ts1. Constants that do have a timestamp
           greater than or equal to ts1 are preserved via a raising of
           v1. As in prune, n is assumed to be the length of the list
           args. The terms in args are assumed to be constants or de
           Bruijn indices. 
        *)
        fun prune_and_raise nil 0 = (false,nil,nil)
          | prune_and_raise ((a as (ref (Term.dB i)))::rargs) n =
              let val (pruned,inds1,inds2) = prune_and_raise rargs (n-1)
              in if (i > lev) 
                 then let val j = bvindex((Term.dB (i-lev)),a1,l1)
                      in if (j = 0) 
                         then (true,inds1,inds2)
                         else (pruned,
                               (ref (Term.dB (j+lev)))::inds1,
                               (ref (Term.dB n))::inds2)
                      end
                 else (pruned,a::inds1,(ref (Term.dB n))::inds2)
              end
          | prune_and_raise ((c as ref (Term.const (_,cts,_)))::rargs) n =
              let val (pruned,inds1,inds2) = prune_and_raise rargs (n-1)
              in if (cts <= ts1) 
                 then (pruned,c::inds1,(ref (Term.dB n))::inds2)
                 else let val i = cindex(!c,a1,l1)
                      in if (i = 0)
                         then (true,inds1,inds2)
                         else (pruned,
                               (ref (Term.dB (i+lev)))::inds1,
                               (ref (Term.dB n))::inds2)
                      end
              end

      in if (ts1 < ts2) 
         then let val (raised,args11,args12) = raise_var a1 l1
                  val (pruned,args21,args22) = prune a2 (length a2)
              in (raised orelse pruned, args11 @ args21, args12 @ args22)
              end
         else (prune_and_raise a2 (length a2))
      end


(* Generating the arguments of a pruning substitution for the case
   when trying to unify two flexible terms of the form (v t1 ... tn)
   and lam ... lam (v s1 ... sm) where there are j abstractions at the 
   head of the second term. The first two arguments to prune_same_var
   are the two lists of arguments, the third argument is j (i.e. the
   number of abstractions at the head) and the last argument is n+j. It
   is assumed that type consistency has been checked beforehand,
   i.e. n+j is indeed equal to m and also that the two argument lists
   are known to be of the form required for LLambda unification. The
   computation essentially does the eta fluffing of the first term on
   the fly (i.e. each ti has to be lifted over j abstractions and and
   j new arguments bound by these abstractions are added to the first
   term). 
*)
fun prune_same_var nil nil _ _ = nil
  | prune_same_var nil ((ref (Term.dB i))::a2) j bl = 
      if (i = j) 
      then (ref (Term.dB bl)::(prune_same_var nil a2 (j-1) (bl-1)))
      else (prune_same_var nil a2 (j-1) (bl-1))
  | prune_same_var nil (_::a2) j bl = (prune_same_var nil a2 (j-1) (bl-1))
  | prune_same_var ((ref (Term.const c1))::a1) 
                   ((ref (Term.const c2))::a2) j bl =
      if (c1 = c2) 
      then ((ref (Term.dB bl))::(prune_same_var a1 a2 j (bl-1)))
      else (prune_same_var a1 a2 j (bl-1))
  | prune_same_var ((ref (Term.dB i1))::a1) ((ref (Term.dB i2))::a2) j bl =
      if ((i1+j) = i2) 
      then ((ref (Term.dB bl))::(prune_same_var a1 a2 j (bl-1)))
      else (prune_same_var a1 a2 j (bl-1))
  | prune_same_var (_::a1) (_::a2) j bl = prune_same_var a1 a2 j (bl-1)


(* Given a term of the form (app (h1,a1)) where h1 is a variable and
   another term t2, generate an LLambda substitution fpr h1 if this is 
   possible, making whatever pruning and raising substitution that are
   necessary to variables appearing within t2. Exceptions can be
   raised from this code if a non LLambda situation is discovered or
   there is failure in unification or a type mismatch (possible if an
   a priori type checking has not been done) is encountered. The
   second argument is assumed to be in head normal form, the first and
   second are assumed to be dereferenced and the last argument is the
   number of arguments in a1.

   The unification computation is split into two parts, one that
   examines the top level structure of t2 and the other that descends
   into its nested subparts. This organization is useful primarily
   because h1, the variable head, of the first term can appear at the
   toplevel in t2 without sacrificing unifiability but not in a nested
   part. 
*)
fun makesubst (h1 as (ref (Term.var (_,ts1,_)))) t2 a1 n =
  let 
    val a1 = (map Norm.hnorm a1)

    (* Generating a substitution term and performing raising and
       pruning substitutions corresponding to a non top-level
       (sub)term. In this case the variable being bound cannot appear
       embedded inside the term. This code assumes that its term
       argument is head normalized and dereferenced. Exceptions can be
       raised if unification fails or if LLambda conditions are found
       to be violated. 
    *)
    fun nested_subst (c as ref (Term.const (_,cts,_))) lev =
         if (cts <= ts1) 
         then c
         else let val j = cindex(!c,a1,n) 
              in if (j <> 0) 
                 then (ref (Term.dB (j+lev))) 
                 else raise OccursCheck
              end
      | nested_subst (bv as ref (Term.dB i)) lev =
         if (i <= lev) then bv
         else let val j = bvindex((Term.dB (i-lev)),a1,n) 
              in if (j <> 0) 
                 then (ref (Term.dB (j+lev))) 
                 else raise OccursCheck
              end
      | nested_subst (v as ref (Term.var (_,ts2,_))) lev =
         if (v = h1) then raise OccursCheck
         else let val (changed,a1',a2') = 
                             (raise_and_invert ts1 ts2 a1 nil lev)
              in if (changed orelse (ts1 < ts2))
                 then let val h' = if (ts1 < ts2) 
                                   then Term.freshvar(ts1)
                                   else Term.freshvar(ts2)
                      in ((Term.bind v (ref (Term.app (h',a2')))); 
                          (ref (Term.app (h',a1'))))
                      end
                 else (ref (Term.app (v,a1')))
              end
      | nested_subst (ref (Term.lam (n,t))) lev = 
         ref (Term.lam (n,(nested_subst (Term.deref t) (lev+n))))
      | nested_subst (ref (Term.app (h2,a2))) lev = 
         (fn (h2' as ref (Term.const _)) => 
                ref (Term.app ((nested_subst h2' lev),
                          (map (fn x => (nested_subst (Norm.hnorm x) lev)) 
                               a2)))
           | (h2' as ref (Term.dB _)) =>
                ref (Term.app ((nested_subst h2' lev),
                          (map (fn x => (nested_subst (Norm.hnorm x) lev)) 
                               a2)))
           | (h2' as ref (Term.var (_,ts2,_))) => 
                if (h2' = h1) then raise OccursCheck
                else let val a2 = (map Norm.hnorm a2)
                     in if (check_flex_args a2 ts2)
                        then 
                          let val (changed,a1',a2') = 
                                  (raise_and_invert ts1 ts2 a1 a2 lev)
                          in 
                            if changed
                            then 
                              let val h' = if (ts1 < ts2)
                                           then Term.freshvar(ts1)
                                           else Term.freshvar(ts2)
                              in ((Term.bind h2' 
                                             (Term.makelambda (length a2) 
                                                  (ref (Term.app (h',a2')))));
                                  (ref (Term.app (h',a1'))))
                              end
                            else 
                              if (ts1 < ts2)
                              then let val h' = Term.freshvar(ts1)
                                   in ((Term.bind h2' h'); 
                                       (ref (Term.app (h',a1))))
                                   end
                              else (ref (Term.app (h2',a1')))
                          end
                        else raise NotLLambda
                     end
         ) (Term.deref h2)


    (* processing toplevel structure in generating a substitution. 
       First descend under abstractions. Then if the term is a
       variable, generate the simple substitution. Alternatively, if
       it is an application with the variable being bound as its head,
       then generate the pruning substitution. In all other cases,
       pass the task on to nested_subst. An optimization is possible
       in the case that the term being examined has no outer
       abstractions (i.e. lev = 0) and its head is a variable with a
       time stamp greater than that of the variable being bound. In
       this case it may be better to invoke raise_and_invert
       directly with the order of the "terms" reversed. 

       The incoming term is assumed to be head normalized and
       dereferenced. 
    *)
    fun toplevel_subst (ref (Term.lam (n,t2))) lev = 
          toplevel_subst (Term.deref t2) (lev+n)
      | toplevel_subst (h2 as ref (Term.var _)) lev =
          if (h1 = h2) 
          then if ((n=0) andalso (lev=0)) then h1 else raise TypesMismatch
          else (Term.makelambda (lev+n) h2)
      | toplevel_subst (t2 as (ref (Term.app (h2,a2)))) lev =
          (fn (h2' as ref (Term.var (_,ts2,_))) =>
               if (h1 = h2') 
               then 
                 let val a2' = (map Norm.hnorm a2)
                 in if (check_flex_args a2' ts2)
                    then
                      let val bindlen = n + lev
                      in if (bindlen = (length a2'))
                         then 
                           let 
                             val h1' = Term.freshvar(ts1)
                             val args = prune_same_var a1 a2' lev bindlen
                           in (Term.makelambda bindlen 
                                               (ref (Term.app (h1',args))))
                           end
                         else raise TypesMismatch
                      end
                    else raise NotLLambda
                 end
               else Term.makelambda (n+lev) (nested_subst t2 lev)
            | _ => Term.makelambda (n+lev) (nested_subst t2 lev)
          ) (Term.deref h2)
      | toplevel_subst t2 lev = Term.makelambda (n+lev) (nested_subst t2 lev)

  in if (check_flex_args a1 ts1) 
     then (toplevel_subst t2 0)
     else raise NotLLambda
  end




(* Unifying the arguments of two rigid terms with the same head, these
   arguments being given as lists. Exceptions are raised if
   unification fails or if there are unequal numbers of arguments; the
   latter will not arise if type checking has been done.
*)
fun unify_list nil nil = ()
  | unify_list (a1::as1) (a2::as2) = 
      (unify (Norm.hnorm a1) (Norm.hnorm a2) ; unify_list as1 as2)
  | unify_list _ _ = raise TypesMismatch

and 

(* unifying a constant with a term that is not a variable or an
   application. If it is a lambda, binders need to be equalized and so
   this becomes an application-term unification problem. 
*)
    unify_const_term (t1 as ref (Term.const c1)) 
                     (t2 as ref (Term.const c2)) =
      if (c1 = c2) then () else raise ConstClash(t1, t2)
  | unify_const_term (t1 as ref (Term.const c1)) (ref (Term.lam (n,t2))) =
     let val a1 = lift_args nil n
     in unify_app_term t1 a1 (ref (Term.app (t1,a1))) t2
     end
  | unify_const_term t1 t2 = raise ConstClash(t1,t2)


and 

(* Unifying a bound variable with a term that is not a variable,
   an application or a constant. If it is a lambda, binders need to be
   equalized and this becomes an application-term unification problem.
*)
    unify_bv_term (t1 as ref (Term.dB n1)) (t2 as ref (Term.dB n2)) =
      if n1 = n2 then () else raise ConstClash(t1, t2)
  | unify_bv_term (t1 as ref (Term.dB c1)) (ref (Term.lam (n,t2))) =
      let val t1' = lift t1 n
          val a1' = lift_args nil n
      in unify_app_term t1' a1' (ref (Term.app (t1',a1'))) t2
      end
  | unify_bv_term t1 t2 = raise ConstClash(t1,t2)

and

(* unifying an application, supplied as the (dereferenced) head and
    arguments list in addition to the term itself, with another
    dereferenced and head normalized term different from a variable. 
*)
    unify_app_term (h1 as (ref (Term.var _))) a1 t1 t2 =
     let val n = (length a1)
     in (Term.bind h1 (makesubst h1 t2 a1 n))
     end
  | unify_app_term (h1 as (ref (Term.const c1)))
                   a1 t1 (t2 as (ref (Term.app (h2,a2)))) =
     (fn (h as (ref (Term.const c2))) => if c1 = c2 
                             then (unify_list a1 a2)
                             else raise ConstClash(h1, h)
       | (h as (ref (Term.dB n2))) => raise ConstClash(h1, h)
       | h2' => let val m = (length a2)
                in (Term.bind h2' (makesubst h2' t1 a2 m))
                end
     ) (Term.deref h2)
  | unify_app_term (h1 as (ref (Term.dB n1)))
                   a1 t1 (t2 as (ref (Term.app (h2,a2)))) =
     (fn (h as (ref (Term.const c2))) => raise ConstClash(h1, h)
       | (h as (ref (Term.dB n2))) => 
             if (n1 = n2) then (unify_list a1 a2) 
             else raise ConstClash(h1,h)
       | h2' => let val m = (length a2)
                in (Term.bind h2' (makesubst h2' t1 a2 m))
                end
     ) (Term.deref h2)
  | unify_app_term h1 a1 t1 (ref (Term.lam (n,t2))) = 
     let val h1' = lift h1 n
         val a1' = lift_args a1 n
         val t1' = (ref (Term.app (h1',a1'))) 
     in unify_app_term h1' a1' t1' t2
     end
  | unify_app_term h1 _ _ h2 =  (* second term must be const or dB *)
     raise ConstClash(h1, h2)

and 

(* The main unification procedure. When called with terms as arguments, it
   either succeeds and realizes the unification substitutions as side effects
   or it raises an exception to indicate nonunifiability or to signal
   a case outside of the LLambda subset. When an exception is raised,
   it is necessary to catch this and at least undo bindings for
   variables made in the attempt to unify. This has not been included
   in the code at present.

   This procedure assumes that the two terms it gets are in
   dereferenced, head normal form and that there are no iterated
   lambdas or applications at the top level. Any necessary adjustment
   of binders through the eta rule is done on the fly. 
*)
    unify (t1 as ref (Term.var (_,ts1,_))) t2 = 
            (Term.bind t1 (makesubst t1 t2 nil 0))
  | unify t1 (t2 as ref (Term.var (_,ts2,_))) = 
            (Term.bind t2 (makesubst t2 t1 nil 0))
  | unify (t1 as ref (Term.app (h1,a1))) t2 = 
                (unify_app_term (Term.deref h1) a1 t1 t2)
  | unify t1 (t2 as ref (Term.app (h2,a2))) = 
                (unify_app_term (Term.deref h2) a2 t2 t1)
  | unify (t1 as ref (Term.const _)) t2 = (unify_const_term t1 t2)
  | unify t1 (t2 as ref (Term.const _)) = (unify_const_term t2 t1)
  | unify (t1 as ref (Term.dB _)) t2 = (unify_bv_term t1 t2)
  | unify t1 (t2 as ref (Term.dB _)) = (unify_bv_term t2 t1)
  | unify (ref (Term.lam (n1,t1))) (ref (Term.lam (n2,t2))) =  
          if (n1 > n2) 
          then (unify (Term.makelambda (n1-n2) t1) (Term.deref t2))
          else (unify (Term.deref t1) (Term.makelambda (n2-n1) t2))


(* The interface to the functionality of this module *)
fun pattern_unify t1 t2 = unify (Norm.hnorm t1) (Norm.hnorm t2)
end
