(****************************************************************************)
(*   The abstract data type for streams using side effects                  *)       
(*           Copyright (C) 2004 Dale Miller                                 *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

signature USTREAM = 
sig
   type 'a elm 
   datatype 'a ustream = 
      empty | ustream of 'a * ('a ustream elm ref)
   
   exception Head
   exception Tail
   val head : 'a ustream -> 'a
   val tail : 'a ustream -> 'a ustream
   val snil : 'a ustream
   val snull : 'a ustream -> bool
   val scons : ('a * (unit -> 'a ustream)) -> 'a ustream
   
   val squash : ('a ustream) ustream -> 'a ustream
   val maps : ('a -> 'b) -> 'a ustream -> 'b ustream
   val mapsl : ('a -> 'b) -> 'a ustream -> 'b list
   val chop : ('a ustream -> bool) * 'a ustream * 'a ustream -> 'a ustream   
end

structure UStream :> USTREAM =
struct
   datatype 'a cell = delayedcell of unit -> 'a |
                      forcedcell of 'a
   type 'a elm = 'a cell
   fun getcell(t as ref(delayedcell t')) =
       let val v = t'() in 
           (t := (forcedcell v); v) end
     | getcell(ref (forcedcell v)) = v
   fun mkcell t = ref(delayedcell t)                         

   datatype 'a ustream = 
      empty | ustream of 'a * ('a ustream elm ref)
      
   exception Head
   exception Tail
   
   fun head empty = raise Head
     | head(ustream(x,_)) = x
   fun tail empty = raise Tail
     | tail(ustream(_,t)) = getcell(t)
   val snil = empty
   fun snull empty = true
     | snull(ustream(_,_)) = false
   fun scons(x,t) = ustream(x,mkcell(t)) 

   (* Take a ustream of ustream and flatten it into a single ustream:  the
      elements of the first stream precede elements of later streams.
   *)
   fun squash s_of_s =
       if snull(s_of_s) then snil else
          let val h=head(s_of_s) in 
             if snull(h) then squash(tail(s_of_s)) else
             scons(head h, fn ()=>squash(scons(tail h,fn () => tail s_of_s)))
          end;

   fun maps f s =
       if snull s then snil else
          scons(f(head s),fn () => maps f (tail s));

   (*  map f over a stream to get a list.  *)
   fun mapsl f s = 
       if (snull s) then nil
          else (f (head s))::(mapsl f (tail s));

   (* chop(pred,s,init) takes the initial portion of the stream s for
      which pred does not hold and appends it to the stream init.
   *)

   fun chop(pred,s,init) =
     if snull s orelse pred s then init
        else scons((head s),fn ()=>chop(pred,(tail s),init));        
end

